#include <hex/data_processor/node.hpp>

namespace hex::plugin::builtin {

    void registerBasicDataProcessorNodes();
    void registerVisualDataProcessorNodes();
    void registerLogicDataProcessorNodes();
    void registerControlDataProcessorNodes();
    void registerDecodeDataProcessorNodes();
    void registerMathDataProcessorNodes();
    void registerOtherDataProcessorNodes();

}