#include <hex/helpers/literals.hpp>

namespace hex::plugin::builtin {

    void drawDemangler();
    void drawASCIITable();
    void drawRegexReplacer();
    void drawColorPicker();
    void drawMathEvaluator();
    void drawGraphingCalculator();
    void drawBaseConverter();
    void drawByteSwapper();
    void drawPermissionsCalculator();
    // void drawFileUploader();
    void drawWikiExplainer();

    void drawIEEE754Decoder();
    void drawInvariantMultiplicationDecoder();
    void drawTCPClientServer();
    void drawEuclidianAlgorithm();

    void drawFileToolShredder();
    void drawFileToolSplitter();
    void drawFileToolCombiner();

    void drawHTTPRequestMaker();

}
